#Fitting Drude model to mid infrared ellipsometry data of gold
import numpy as np
import matplotlib.pyplot as plt
#from matplotlib.pyplot import plot, ion, show
import time
import math
import lmfit 

######################
#USER INPUT
index0=1  #Index of refraction of the ambient
LowEnCut=200   #selecting LOW energy cut range
HighEnCut=1100	  #selecting HIGH energy cut range

file1 = open('131121_LiF_MIR_elli_80.dat', 'r')
#file1 = open('131121_LiF_MIR_elli.dat', 'r')
#file1 = open('GoldMirror2_MIR80deg.dat', 'r')

#Values of parameters are loaded from ParamsInput.dat


#END OF USER INPUT
######################

####################
#Auto loading of values of parameters from  ParamsInput.dat
#Setting values of parameters of dielectric function model
NumberOscilators= {"eINF": 0 , "Lorentz": 0}   #dictionary that contains type of all models and according number of oscillators. 
params = lmfit.Parameters()  #class of parameters for fitting
#adding constant (epsilon "infinity")


def IsFitted(Str): # procedura na detekci zda se dany parametr fituje pomoci znaku "f"
	if (Str[0:1]=="F") or (Str[0:1]=="f"):
		Fit=False
		StrRed=float(Str[1:])
	else:
		Fit=True
		StrRed=float(Str)
	return 	Fit,StrRed  #vraci boolean Fit, ktery je false, jestli je to fixvany parametr,  a StrRed je hodnota parametru

f = open('ParamsInput.dat', 'r')  #otevreni souboru se vstupnimi parametry
lines = f.readlines()    #nacteni souboru se vstupnimi parametry
lines = [x.strip('\t') for x in lines] # ocisteni od koncovych tabulatoru
for i in range(len(lines)):
	line=lines[i].split()   #rozdeleni radku na jednotive stringy
	if len(line)<2: break  #pokud je prazdna radka, tak konci cteni, za ni muzou byt nepouzite oscilatory
	if line[0]=='DataType':  #nacteni typu data, jako Raman, R (odrazivost) atp. 
		NumberOscilators["DataType"]= line[1]
	if line[0]=='eINF':  #zarazeni konstanty eINF
		NumberOscilators["eINF"]=NumberOscilators["eINF"]+1        # zvysuje pocet daneho oscilatoru v prehledu oscilatoru 
		Fitted,Val=IsFitted(line[1])        #nacteni hodnoty a osetreni znaku "f"
		params.add('eINF', value=Val,vary=Fitted)     #generace parametru, jeho hodnoty a toho, zda se fituje
	if line[0]=='Lorentz': #zarazeni Lorenztova oscilatoru
		NumberOscilators["Lorentz"]=NumberOscilators["Lorentz"]+1    # zvysuje pocet daneho oscilatoru v prehledu oscilatoru 
		OscNum=NumberOscilators["Lorentz"]        
		Fitted,Val=IsFitted(line[1])        
		params.add('PlasmaEn'+str(OscNum), value=Val,vary=Fitted)
		Fitted,Val=IsFitted(line[2])
		params.add('Energy'+str(OscNum), value=Val,vary=Fitted)
		Fitted,Val=IsFitted(line[3])
		params.add('Gamma'+str(OscNum), value=Val,vary=Fitted)

#End of Auto loading of values of parameters from  ParamsInput.dat
####################






Lines = file1.readlines()   #Reading data file into list of Lines
Columns=Lines[2].split()	
Units=Columns[0]
print("DataUnits=",Units)
AngleList=[]  #Angle of incidence
EnergyList=[]  #Energy of radiation
PsiList=[]	#Psi
dPsiList=[]  #error of Psi
DeltaList=[]
dDeltaList=[]
#reading of WVASE dat file
for n in range(4,len(Lines)):
	m=0
	DatFile="Unknown"
	try:
		float(Lines[n][0])  #trying converting the first column to float, this is the syntax of VASE dat files
		DatFile="VASE"
		m=0
	except: #IRVASE dat file convention (starting with "E" column)
		if (Lines[n][0]=="E"): 
			m=1
			DatFile="IRVASE"
	if (DatFile=="VASE") or (DatFile=="IRVASE"):
		Columns=Lines[n].split()			
		En=float(Columns[0+m])
		if (En>LowEnCut) and  (En<HighEnCut):   # if energy falls within selected range, append data
			EnergyList.append(float(Columns[0+m]))
			AngleList.append(float(Columns[1+m]))
			PsiList.append(float(Columns[2+m]))
			DeltaList.append(float(Columns[3+m]))
			dPsiList.append(float(Columns[4+m]))
			dDeltaList.append(float(Columns[5+m]))

class ExperimentalData:  #definition of class of all experimental data, possible to add more (thickness, depolarization...)
	Angle=np.array(AngleList)  #creating NUMPY arrays from lists
	Energy=np.array(EnergyList)  
	Psi=np.array(PsiList)	
	dPsi=np.array(dPsiList)  
	Delta=np.array(DeltaList)
	dDelta=np.array(dDeltaList)
Data=ExperimentalData   #creating Data variable where all experimental data are stored

#Data.Energy=Data.Energy/8066  #converting cm-1 to eV, if desired

def Lorentzian(x,PlasmaEn,En0,Gamma):  #Lorentz model
    return PlasmaEn/(En0**2-x**2-1j*Gamma*x) 

#Definitions of Fresnel amplitudes 
def costheta(ctheta0, cindex0, cindex):	#Complex angles: thata0= angle of incidence, index0 and index: index of refr. of incident material and refractive material 
	return (1-pow(cindex0/cindex*np.sin(ctheta0),2))**0.5
def rp(rindex1,rindex2,rcos1,rcos2): # Fresnel rp coeficient 
	return (rindex2*rcos1-rindex1*rcos2)/(rindex2*rcos1+rindex1*rcos2)
def rs(sindex1,sindex2,scos1,scos2):  # Fresnel rs coeficient Function 
	return (sindex1*scos1-sindex2*scos2)/(sindex1*scos1+sindex2*scos2)

def ResidualModelFunction(params, NumberOscilators, Energy, theta0):
	theta0=np.radians(theta0)  #conversion to radians
	v = params.valuesdict()    #unpacking all parameters to v
	#Building the model  dielectric function
	eps=np.full(len(Energy),0.0)  #difinition of nd array for dielectric function
	for n in range(NumberOscilators["eINF"]):  eps = eps+ v['eINF']   # adding the eps_infinity constant
	for n in range(NumberOscilators["Lorentz"]): #adding Lorentz oscillator
		eps=eps+Lorentzian(Energy,v["PlasmaEn"+str(n+1)],v["Energy"+str(n+1)],v["Gamma"+str(n+1)])        
	index1 = eps**0.5              # index of refraction from dielectric function
	costheta1 = costheta(theta0,index0,index1) #complex angle inside of material
	#ellipsometric rho =rp/rs
	rho=rp(index0,index1,np.cos(theta0),costheta1)/rs(index0,index1,np.cos(theta0),costheta1)
	#below is definition of Delta in  Woollam rotating compensator convention, Woolam delta is between -90 and 270deg, the python delta is between -180 and 180deg
	DeltaTheory=-np.angle(rho*1j)+math.pi/2  
	PsiTheory=np.arctan(np.abs(rho))    
	return np.rad2deg(PsiTheory), np.rad2deg(DeltaTheory), eps

def ResidualModel(params, NumberOscilators, Data):
	PsiTheory, DeltaTheory, eps= ResidualModelFunction(params, NumberOscilators, Data.Energy, Data.Angle)
	#calculatin apropriate residual function: here sum of the queares of psi and delta differences to the measured data weighted by errorbars
	Resid= pow((PsiTheory-Data.Psi)/Data.dPsi,2)+ pow((DeltaTheory-Data.Delta)/Data.dDelta,2)
	return Resid   


print ("Start of  fitting")
TimeStart=time.time()
minner = lmfit.Minimizer(ResidualModel, params, fcn_args=(NumberOscilators, Data))  
result = minner.minimize()  #zde se uchovovaji vsechny vysledky fitovani
ResultParams= result.params   #resulting parameters from the fit
TimeEnd=time.time()    
print ("Time of fitting: ", TimeEnd-TimeStart , " s")


#calculation of starting model spectra 
PsiStart, DeltaStart, epsStart= ResidualModelFunction(params, NumberOscilators, Data.Energy, Data.Angle)
#calculation fitted model spectra 
PsiTheory, DeltaTheory, epsTheory= ResidualModelFunction(ResultParams, NumberOscilators, Data.Energy, Data.Angle)

#vytisteni statistiky na obrazovku a do souboru
print(lmfit.fit_report(result,min_correl=0.5))   
FileFit = open('Statistics.dat', 'w+')
print  (lmfit.fit_report(result,min_correl=0.5), file=FileFit)   
FileFit.close()

def SaveToFile(NameOfFile,Col1,Col2, HeaderCol2, HeaderCol3):
	SaveFile = open(NameOfFile, 'w')
	SaveLine = "Energy"+"\t" + HeaderCol2 +"\t" + HeaderCol3
	print(SaveLine,file=SaveFile)
	for i in range(len(Data.Energy)):
		SaveLine=str("{0:.4g}".format(Data.Energy[i]))
		SaveLine=SaveLine + "\t" + str("{0:.4g}".format(Col1[i]))
		SaveLine=SaveLine + "\t" +  str("{0:.4g}".format(Col2[i]))
		print(SaveLine,file=SaveFile)
	SaveFile.close()
	
SaveToFile("PsiDeltaFit.dat", PsiTheory, DeltaTheory, "PsiFit", "DeltaFit")
SaveToFile("DielFunct.dat", np.real(epsTheory), np.imag(epsTheory), "e1Fit", "e2fit")
if (Units=="eV"): OptConductivity=-1j/60*Data.Energy*(epsTheory-1)*8066
if (Units=="1/cm"): OptConductivity=-1j/60*Data.Energy*(epsTheory-1)
SaveToFile("OptConduct.dat", np.real(OptConductivity), np.imag(OptConductivity), "s1Fit", "s2fit")


#Saving of resulting parameters into ParamsOutput.dat
FileParOut = open('ParamsOutput.dat', 'w')
OutParams=result.params
def StrForm(strval):    #funkce na formatovani vystupniho cisla na 4 platne cislice a prevod do stringu
	return str("{0:.4g}".format(strval))
print ("DataType\t" + NumberOscilators["DataType"], file=FileParOut)
for i in range(NumberOscilators["eINF"]): 
        word="eINF"+"\t"
        if OutParams["eINF"].vary==False:word=word+"f"          #pokud je parametr fixovany, priadava "f"
        word=word+StrForm(OutParams["eINF"].value)
        print (word, file=FileParOut)
for i in range(NumberOscilators["Lorentz"]): 
        word="Lorentz"+"\t"
        if OutParams['PlasmaEn'+str(i+1)].vary==False:word=word+"f"
        word=word+StrForm(OutParams['PlasmaEn'+str(i+1)].value)+"\t"
        if OutParams['Energy'+str(i+1)].vary==False:word=word+"f"
        word=word+StrForm(OutParams['Energy'+str(i+1)].value)+"\t"
        if OutParams['Gamma'+str(i+1)].vary==False:word=word+"f"
        word=word+StrForm(OutParams['Gamma'+str(i+1)].value)
        print (word, file=FileParOut)

#plotting data and theoretical spectra
fig, (ax1, ax2) = plt.subplots(1, 2)	#definition of the figure
ax1.plot(Data.Energy, Data.Psi, 'b-',label='PsiExp')	
ax2.plot(Data.Energy, Data.Delta, 'b-',label='DeltaExp')
ax1.plot(Data.Energy, PsiStart, 'g-',label='PsiTheory Start')	
ax2.plot(Data.Energy, DeltaStart, 'g-',label='DeltaTheory Start')	
ax1.plot(Data.Energy, PsiTheory, 'r-',label='PsiTheory Fit')	
ax2.plot(Data.Energy, DeltaTheory, 'r-',label='DeltaTheory Fit')	
ax1.set_title('Psi')
ax2.set_title('Delta')
plt.legend(loc='upper right', shadow=False)
plt.show()